@@ -21,7 +21,8 @@ use std::i32;
 use std::sync::Mutex;
 
 #[derive(Default)]
-struct NegotiationInfos {
+struct State {
+    decoder: dav1d::Decoder,
     input_state:
         Option<gst_video::VideoCodecState<'static, gst_video::video_codec_state::Readable>>,
     output_info: Option<gst_video::VideoInfo>,
@@ -30,8 +31,7 @@ struct NegotiationInfos {
 
 #[derive(Default)]
 pub struct Dav1dDec {
-    decoder: Mutex<dav1d::Decoder>,
-    negotiation_infos: Mutex<NegotiationInfos>,
+    state: Mutex<Option<State>>,
 }
 
 static CAT: Lazy<gst::DebugCategory> = Lazy::new(|| {
@@ -45,19 +45,29 @@ static CAT: Lazy<gst::DebugCategory> = Lazy::new(|| {
 impl Dav1dDec {
     pub fn gst_video_format_from_dav1d_picture(
         &self,
+        element: &super::Dav1dDec,
         pic: &dav1d::Picture,
     ) -> gst_video::VideoFormat {
         let bpc = pic.bits_per_component();
         let format_desc = match (pic.pixel_layout(), bpc) {
-            // (dav1d::PixelLayout::I400, Some(dav1d::BitsPerComponent(8))) => "GRAY8",
+            (dav1d::PixelLayout::I400, Some(dav1d::BitsPerComponent(8))) => "GRAY8",
+            #[cfg(target_endian = "little")]
+            (dav1d::PixelLayout::I400, Some(dav1d::BitsPerComponent(16))) => "GRAY16_LE",
+            #[cfg(target_endian = "big")]
+            (dav1d::PixelLayout::I400, Some(dav1d::BitsPerComponent(16))) => "GRAY16_BE",
             // (dav1d::PixelLayout::I400, Some(dav1d::BitsPerComponent(10))) => "GRAY10_LE32",
-            (dav1d::PixelLayout::I400, _) => return gst_video::VideoFormat::Unknown,
             (dav1d::PixelLayout::I420, _) => "I420",
             (dav1d::PixelLayout::I422, Some(dav1d::BitsPerComponent(8))) => "Y42B",
             (dav1d::PixelLayout::I422, _) => "I422",
             (dav1d::PixelLayout::I444, _) => "Y444",
-            (dav1d::PixelLayout::Unknown, _) => {
-                gst_warning!(CAT, "Unsupported dav1d format");
+            (layout, bpc) => {
+                gst_warning!(
+                    CAT,
+                    obj: element,
+                    "Unsupported dav1d format {:?}/{:?}",
+                    layout,
+                    bpc
+                );
                 return gst_video::VideoFormat::Unknown;
             }
         };
@@ -81,7 +91,7 @@ impl Dav1dDec {
             }
         };
         f.parse::<gst_video::VideoFormat>().unwrap_or_else(|_| {
-            gst_warning!(CAT, "Unsupported dav1d format: {}", f);
+            gst_warning!(CAT, obj: element, "Unsupported dav1d format: {}", f);
             gst_video::VideoFormat::Unknown
         })
     }
@@ -92,9 +102,11 @@ impl Dav1dDec {
         pic: &dav1d::Picture,
         format: gst_video::VideoFormat,
     ) -> Result<(), gst::FlowError> {
+        let mut state_guard = self.state.lock().unwrap();
+        let state = state_guard.as_mut().unwrap();
+
         let negotiate = {
-            let negotiation_infos = self.negotiation_infos.lock().unwrap();
-            match negotiation_infos.output_info {
+            match state.output_info {
                 Some(ref i) => {
                     (i.width() != pic.width())
                         || (i.height() != pic.height() || (i.format() != format))
@@ -112,32 +124,37 @@ impl Dav1dDec {
             pic.width(),
             pic.height()
         );
-        let output_state = {
-            let negotiation_infos = self.negotiation_infos.lock().unwrap();
-            let input_state = negotiation_infos.input_state.as_ref();
-            element.set_output_state(format, pic.width(), pic.height(), input_state)
-        }?;
+
+        let input_state = state.input_state.as_ref().cloned();
+        drop(state_guard);
+
+        let output_state =
+            element.set_output_state(format, pic.width(), pic.height(), input_state.as_ref())?;
         element.negotiate(output_state)?;
         let out_state = element.output_state().unwrap();
-        {
-            let mut negotiation_infos = self.negotiation_infos.lock().unwrap();
-            negotiation_infos.output_info = Some(out_state.info());
-        }
+
+        let mut state_guard = self.state.lock().unwrap();
+        let state = state_guard.as_mut().unwrap();
+        state.output_info = Some(out_state.info());
 
         Ok(())
     }
 
-    fn flush_decoder(&self) {
-        let decoder = self.decoder.lock().unwrap();
-        decoder.flush();
+    fn flush_decoder(&self, _element: &super::Dav1dDec) {
+        let mut state_guard = self.state.lock().unwrap();
+        let state = state_guard.as_mut().unwrap();
+        state.decoder.flush();
     }
 
     fn decode(
         &self,
+        element: &super::Dav1dDec,
         input_buffer: &gst::BufferRef,
         frame: &gst_video::VideoCodecFrame,
     ) -> Result<Vec<(dav1d::Picture, gst_video::VideoFormat)>, gst::FlowError> {
-        let mut decoder = self.decoder.lock().unwrap();
+        let mut state_guard = self.state.lock().unwrap();
+        let state = state_guard.as_mut().unwrap();
+
         let timestamp = frame.dts().map(|ts| *ts as i64);
         let duration = frame.duration().map(|d| *d as i64);
 
@@ -145,16 +162,27 @@ impl Dav1dDec {
         let input_data = input_buffer
             .map_readable()
             .map_err(|_| gst::FlowError::Error)?;
-        let pictures = decoder
-            .decode(input_data, frame_number, timestamp, duration, || {})
-            .map_err(|e| {
-                gst_error!(CAT, "Decoding failed (error code: {})", e);
-                gst::FlowError::Error
-            })?;
+        let pictures =
+            match state
+                .decoder
+                .decode(input_data, frame_number, timestamp, duration, || {})
+            {
+                Ok(pictures) => pictures,
+                Err(err) => {
+                    gst_error!(CAT, "Decoding failed (error code: {})", err);
+                    return gst_video::video_decoder_error!(
+                        element,
+                        1,
+                        gst::StreamError::Decode,
+                        ["Decoding failed (error code {})", err]
+                    )
+                    .map(|_| vec![]);
+                }
+            };
 
         let mut decoded_pictures = vec![];
         for pic in pictures {
-            let format = self.gst_video_format_from_dav1d_picture(&pic);
+            let format = self.gst_video_format_from_dav1d_picture(element, &pic);
             if format != gst_video::VideoFormat::Unknown {
                 decoded_pictures.push((pic, format));
             } else {
@@ -166,6 +194,7 @@ impl Dav1dDec {
 
     pub fn decoded_picture_as_buffer(
         &self,
+        element: &super::Dav1dDec,
         pic: &dav1d::Picture,
         output_state: gst_video::VideoCodecState<gst_video::video_codec_state::Readable>,
     ) -> Result<gst::Buffer, gst::FlowError> {
@@ -173,29 +202,40 @@ impl Dav1dDec {
         let mut strides = vec![];
         let mut acc_offset: usize = 0;
 
-        let video_meta_supported = self.negotiation_infos.lock().unwrap().video_meta_supported;
+        let mut state_guard = self.state.lock().unwrap();
+        let state = state_guard.as_mut().unwrap();
+        let video_meta_supported = state.video_meta_supported;
+        drop(state_guard);
 
         let info = output_state.info();
         let mut out_buffer = gst::Buffer::new();
         let mut_buffer = out_buffer.get_mut().unwrap();
 
-        // FIXME: For gray support we would need to deal only with the Y component.
-        assert!(info.is_yuv());
-        for component in [
-            dav1d::PlanarImageComponent::Y,
-            dav1d::PlanarImageComponent::U,
-            dav1d::PlanarImageComponent::V,
-        ]
-        .iter()
-        {
-            let dest_stride: u32 = info.stride()[*component as usize].try_into().unwrap();
-            let plane = pic.plane(*component);
-            let (src_stride, height) = pic.plane_data_geometry(*component);
+        let components = if info.is_yuv() {
+            const YUV_COMPONENTS: [dav1d::PlanarImageComponent; 3] = [
+                dav1d::PlanarImageComponent::Y,
+                dav1d::PlanarImageComponent::U,
+                dav1d::PlanarImageComponent::V,
+            ];
+            &YUV_COMPONENTS[..]
+        } else if info.is_gray() {
+            const GRAY_COMPONENTS: [dav1d::PlanarImageComponent; 1] =
+                [dav1d::PlanarImageComponent::Y];
+
+            &GRAY_COMPONENTS[..]
+        } else {
+            unreachable!();
+        };
+        for &component in components {
+            let dest_stride: u32 = info.stride()[component as usize].try_into().unwrap();
+            let plane = pic.plane(component);
+            let (src_stride, height) = pic.plane_data_geometry(component);
             let mem = if video_meta_supported || src_stride == dest_stride {
                 gst::Memory::from_slice(plane)
             } else {
                 gst_trace!(
                     gst::CAT_PERFORMANCE,
+                    obj: element,
                     "Copying decoded video frame component {:?}",
                     component
                 );
@@ -260,7 +300,7 @@ impl Dav1dDec {
             .expect("Output state not set. Shouldn't happen!");
         let offset = pic.offset() as i32;
         if let Some(mut frame) = element.frame(offset) {
-            let output_buffer = self.decoded_picture_as_buffer(pic, output_state)?;
+            let output_buffer = self.decoded_picture_as_buffer(element, pic, output_state)?;
             frame.set_output_buffer(output_buffer);
             element.finish_frame(frame)?;
         } else {
@@ -270,21 +310,26 @@ impl Dav1dDec {
         self.forward_pending_pictures(element)
     }
 
-    fn drop_decoded_pictures(&self) {
-        let mut decoder = self.decoder.lock().unwrap();
-        while let Ok(pic) = decoder.get_picture() {
-            gst_debug!(CAT, "Dropping picture");
+    fn drop_decoded_pictures(&self, element: &super::Dav1dDec) {
+        let mut state_guard = self.state.lock().unwrap();
+        let state = state_guard.as_mut().unwrap();
+
+        while let Ok(pic) = state.decoder.get_picture() {
+            gst_debug!(CAT, obj: element, "Dropping picture");
             drop(pic);
         }
     }
 
     fn pending_pictures(
         &self,
+        element: &super::Dav1dDec,
     ) -> Result<Vec<(dav1d::Picture, gst_video::VideoFormat)>, gst::FlowError> {
-        let mut decoder = self.decoder.lock().unwrap();
+        let mut state_guard = self.state.lock().unwrap();
+        let state = state_guard.as_mut().unwrap();
+
         let mut pictures = vec![];
-        while let Ok(pic) = decoder.get_picture() {
-            let format = self.gst_video_format_from_dav1d_picture(&pic);
+        while let Ok(pic) = state.decoder.get_picture() {
+            let format = self.gst_video_format_from_dav1d_picture(element, &pic);
             if format == gst_video::VideoFormat::Unknown {
                 return Err(gst::FlowError::NotNegotiated);
             }
@@ -297,7 +342,7 @@ impl Dav1dDec {
         &self,
         element: &super::Dav1dDec,
     ) -> Result<gst::FlowSuccess, gst::FlowError> {
-        for (pic, format) in self.pending_pictures()? {
+        for (pic, format) in self.pending_pictures(element)? {
             self.handle_picture(element, &pic, format)?;
         }
         Ok(gst::FlowSuccess::Ok)
@@ -306,12 +351,16 @@ impl Dav1dDec {
 
 fn video_output_formats() -> Vec<glib::SendValue> {
     let values = [
-        // gst_video::VideoFormat::Gray8,
+        gst_video::VideoFormat::Gray8,
+        #[cfg(target_endian = "little")]
+        gst_video::VideoFormat::Gray16Le,
+        #[cfg(target_endian = "big")]
+        gst_video::VideoFormat::Gray16Be,
+        // #[cfg(target_endian = "little")]
+        // gst_video::VideoFormat::Gray10Le32,
         gst_video::VideoFormat::I420,
         gst_video::VideoFormat::Y42b,
         gst_video::VideoFormat::Y444,
-        // #[cfg(target_endian = "little")]
-        // gst_video::VideoFormat::Gray10Le32,
         #[cfg(target_endian = "little")]
         gst_video::VideoFormat::I42010le,
         #[cfg(target_endian = "little")]
@@ -367,7 +416,14 @@ impl ElementImpl for Dav1dDec {
 
     fn pad_templates() -> &'static [gst::PadTemplate] {
         static PAD_TEMPLATES: Lazy<Vec<gst::PadTemplate>> = Lazy::new(|| {
-            let sink_caps = gst::Caps::builder("video/x-av1").build();
+            let sink_caps = if gst::version() >= (1, 19, 0, 0) {
+                gst::Caps::builder("video/x-av1")
+                    .field("stream-format", "obu-stream")
+                    .field("alignment", gst::List::new(["frame", "tu"]))
+                    .build()
+            } else {
+                gst::Caps::builder("video/x-av1").build()
+            };
             let sink_pad_template = gst::PadTemplate::new(
                 "sink",
                 gst::PadDirection::Sink,
@@ -406,24 +462,39 @@ impl ElementImpl for Dav1dDec {
 impl VideoDecoderImpl for Dav1dDec {
     fn start(&self, element: &Self::Type) -> Result<(), gst::ErrorMessage> {
         {
-            let mut infos = self.negotiation_infos.lock().unwrap();
-            infos.output_info = None;
+            let mut state_guard = self.state.lock().unwrap();
+            *state_guard = Some(State {
+                decoder: dav1d::Decoder::new(),
+                input_state: None,
+                output_info: None,
+                video_meta_supported: false,
+            });
         }
 
         self.parent_start(element)
     }
 
+    fn stop(&self, element: &Self::Type) -> Result<(), gst::ErrorMessage> {
+        {
+            let mut state_guard = self.state.lock().unwrap();
+            *state_guard = None;
+        }
+
+        self.parent_stop(element)
+    }
+
     fn set_format(
         &self,
         element: &Self::Type,
-        state: &gst_video::VideoCodecState<'static, gst_video::video_codec_state::Readable>,
+        input_state: &gst_video::VideoCodecState<'static, gst_video::video_codec_state::Readable>,
     ) -> Result<(), gst::LoggableError> {
         {
-            let mut infos = self.negotiation_infos.lock().unwrap();
-            infos.input_state = Some(state.clone());
+            let mut state_guard = self.state.lock().unwrap();
+            let state = state_guard.as_mut().unwrap();
+            state.input_state = Some(input_state.clone());
         }
 
-        self.parent_set_format(element, state)
+        self.parent_set_format(element, input_state)
     }
 
     fn handle_frame(
@@ -432,7 +503,7 @@ impl VideoDecoderImpl for Dav1dDec {
         frame: gst_video::VideoCodecFrame,
     ) -> Result<gst::FlowSuccess, gst::FlowError> {
         let input_buffer = frame.input_buffer().expect("frame without input buffer");
-        for (pic, format) in self.decode(input_buffer, &frame)? {
+        for (pic, format) in self.decode(element, input_buffer, &frame)? {
             self.handle_picture(element, &pic, format)?;
         }
 
@@ -441,21 +512,21 @@ impl VideoDecoderImpl for Dav1dDec {
 
     fn flush(&self, element: &Self::Type) -> bool {
         gst_info!(CAT, obj: element, "Flushing");
-        self.flush_decoder();
-        self.drop_decoded_pictures();
+        self.flush_decoder(element);
+        self.drop_decoded_pictures(element);
         true
     }
 
     fn drain(&self, element: &Self::Type) -> Result<gst::FlowSuccess, gst::FlowError> {
         gst_info!(CAT, obj: element, "Draining");
-        self.flush_decoder();
+        self.flush_decoder(element);
         self.forward_pending_pictures(element)?;
         self.parent_drain(element)
     }
 
     fn finish(&self, element: &Self::Type) -> Result<gst::FlowSuccess, gst::FlowError> {
         gst_info!(CAT, obj: element, "Finishing");
-        self.flush_decoder();
+        self.flush_decoder(element);
         self.forward_pending_pictures(element)?;
         self.parent_finish(element)
     }
@@ -465,9 +536,13 @@ impl VideoDecoderImpl for Dav1dDec {
         element: &Self::Type,
         query: gst::query::Allocation<&mut gst::QueryRef>,
     ) -> Result<(), gst::LoggableError> {
-        self.negotiation_infos.lock().unwrap().video_meta_supported = query
-            .find_allocation_meta::<gst_video::VideoMeta>()
-            .is_some();
+        {
+            let mut state_guard = self.state.lock().unwrap();
+            let state = state_guard.as_mut().unwrap();
+            state.video_meta_supported = query
+                .find_allocation_meta::<gst_video::VideoMeta>()
+                .is_some();
+        }
 
         self.parent_decide_allocation(element, query)
     }
